#ifndef _ARC_H // -*-C++-*-
#define _ARC_H

#include <iostream.h>
#include <libtk/avlcpp.h>
#include "base.h"

class VertexAVL;

class ArcVertex : public Vertex
{
    friend class VertexAVL;
    VertexAVL &m_tree;
    int m_count;

    ArcVertex( Vertex &,  VertexAVL & );

public:
    ~ArcVertex();
    int increment()  { return ++m_count; }
    int decrement()  { return --m_count; }
    int count()      { return m_count; }
    void deleteMaybe();
};

class Arc
{
    ArcVertex *m_v1;
    ArcVertex *m_v2;
    Vertex *m_normal;
    double m_length2;  // of the chord
    void construct( ArcVertex *, ArcVertex * );
    int between( Vertex & );

public:
    Arc( ArcVertex *, ArcVertex * );
    Arc( ArcVertex &, ArcVertex & );
    Arc( const Arc & );
    ~Arc();
    ArcVertex &getV1() { return *m_v1; }
    ArcVertex &getV2() { return *m_v2; }
    Vertex    &getGreatCircle();    // right-handed
    int        inOrOut( Vertex & ); // 1: inside, 0: on line, -1: outside

    static Vertex intersection( Vertex &, Arc & );
    static Vertex intersection( Arc &, Arc & );
    static int    coincident( Arc &, Arc & );
};


class VertexAVL : public AVL<Vertex>
{
public:
    ArcVertex *getVertex( Vertex & );
};


class Triangle
{
    DArray<Arc> m_arcs;

public:
    Triangle( Arc *, Arc *, Arc * );
    ~Triangle();
    void dismantle();   // call this to destroy but preserve arcs
    DArray<Arc> &arcs() { return m_arcs; }
    Arc &getArc( int i ) { return *m_arcs[i]; }
    ArcVertex &getVertex( int i ) { return m_arcs[i]->getV1(); }
    Vertex normal();    // this is expensive!

    static DArray<Triangle> triangulate( DArray<ArcVertex> & );
    static DArray<Triangle> triangulate( DArray<Arc> & );  // arcs consumed!

    friend ostream &operator<<( ostream &, Triangle & );
};

#endif // _ARC_H
