#include <iostream.h>
#include "bspcpp.h"


extern "C" 
{
#include <bsp.h>
}

static FACE    *Polygon2FACE( Polygon * );
static Polygon *FACE2Polygon( FACE * );
static void     renderFaces( BSP *tree, FACE *faceList );
static int      countFaces( BSPNODE * );

BSP::BSP( DArray<Polygon> &array, int f2b )
{
    FACE *faceList = 0;
    FACE *last = 0;

    m_f2b = f2b;
    for( int i=0; i<array.length(); i++ )
    {
	Polygon *p = array[i];
	FACE *face = Polygon2FACE( p );
	if( !last )
	    faceList = face;
	else
	    last->fnext = face;
	last = face;
    }
    last->fnext = 0;
    m_root = BSPconstructTree( &faceList );

    /*
    cout << "BSP::BSP: " << countFaces( (BSPNODE *) m_root )
	<< " faces in tree\n.";
    */
}


BSP::~BSP()
{
    BSPfreeTree( (BSPNODE **) &m_root );
}


DArray<Polygon> BSP::traverse( Vertex &v )
{
    POINT p;
    p.xx = v[0];
    p.yy = v[1];
    p.zz = v[2];
    m_output.clear();
    BSPtraverseTreeAndRender( (BSPNODE *)m_root, &p,
	(RENDERFUNC) renderFaces, this, m_f2b );
    return m_output;
}


void BSP::renderCB( DArray<Polygon> &faces )
{
    m_output.append( faces );
}


FACE *Polygon2FACE( Polygon *p )
{
    DArray<Vertex> &vList = p->getVerts();
    VERTEX *newV = 0;
    VERTEX *last = 0;
    COLOR c;  // won't worry about initializing
    PLANE pl;

    // woah! hack alert!
    int *label = (int *) &c;

    // convert vertex lists
    for( int i=0; i<p->numVerts(); i++ )
    {
	Vertex *oldV = vList[i];
	VERTEX *v = allocVertex( oldV->v[0], oldV->v[1], oldV->v[2] );
	if( !last )
	    newV = v;
	else
	    last->vnext = v;
	last = v;
    }
    last->vnext = allocVertex( vList[0]->v[0], vList[0]->v[1], vList[0]->v[2] );
    last->vnext->vnext = 0;

    *label = p->label();
    Vertex &v0 = *vList[0]; Vertex &v1 = *vList[1]; Vertex &v2 = *vList[2];
    computePlane( v0[0], v0[1], v0[2], v1[0], v1[1], v1[2], v2[0], v2[1], v2[2],
	&pl );
    return allocFace( &c, newV, &pl );
}


Polygon *FACE2Polygon( FACE *f )
{
    DArray<Vertex> array;  // delete when done

    // woah! hack alert!
    int label = *(int *) &(f->color);

    for( VERTEX *v=f->vhead; v->vnext; v=v->vnext ) // skip last one (duplicate)
	array.add( new Vertex(v->xx,v->yy,v->zz) );
    Polygon *p = new Polygon( array, label );
    array.clear( 1 );  // delete vertices
    return p;
}


void renderFaces( BSP *tree, FACE *faceList )
{
    if( !faceList )
	return;
    FACE *f;
    DArray<Polygon> p;

    for( f=faceList; f; f=f->fnext )
	p.add( FACE2Polygon(f) );
    tree->renderCB( p );
}


int countFaces( BSPNODE *node )
{
    int total = 0;
    FACE *f;

    if( !node->node )
	return 0;
    for( f=node->node->sameDir; f; f=f->fnext )
	total++;
    for( f=node->node->oppDir; f; f=f->fnext )
	total++;

    total += countFaces( node->node->negativeSide );
    total += countFaces( node->node->positiveSide );
    return total;
}





