#include <iostream.h>
#include <GL/glut.h>
#include <GL/gl.h>
#include <libtk/window.h>
#include <string.h>
#include "hemi.h"


// the following should probably be used in conjuction
// #define USE_ITEM_BUFFER
// #define USE_SINGLE_WINDOW

const double HCSIZE = 0.1; // length of one edge of hemi cube
const int BORDER = 25;

static void convert( GLubyte *, int );
static void convert( int &, GLubyte * );

class BWWindow : public Window
{
    HemiCube *m_h;

protected:
    void draw() { m_h->draw(); }
    void keyAction( unsigned char key, int x, int y )
	{ if( key==27 ) m_h->compute(); }

public:
    BWWindow( int x, int y, int width, int height, const char *title, 
	      HemiCube *h ) : Window( x, y, width, height, title ) { m_h = h; }
};



HemiCube::HemiCube( Camera &cam, DArray<Polygon> &src, 
		    DArray<Polygon> &occ, double near, double far, int hide )
    : m_sources(src), m_occluders(occ)
{
    m_cam = cam;
    m_hide = hide;
    m_size = cam.m_width;
    m_near = ( near > 0 ) ? near : HCSIZE/2;
    m_far = ( far > 0 ) ? far : 100;
}


void HemiCube::formFactor()
{
    // initialized glut
    char *argv[] = { "ff", 0 };
    int argc = 1;
    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
    glEnable( GL_DEPTH_TEST );
    glDepthFunc( GL_LESS );
    glShadeModel( GL_FLAT );
    glDisable( GL_LIGHTING );
    glCullFace( GL_BACK );


    // make the windows
    int s = m_size;
    int o = s + BORDER;
    m_up = new BWWindow( o, o, s, s, "Up", this );
    m_north = new BWWindow( o, 0, s, s/2, "North", this );
    m_east = new BWWindow( 2*o, o, s/2, s, "East", this );
    m_south = new BWWindow( o, 2*o, s, s/2, "South", this );
    m_west = new BWWindow( 0, o, s/2, s, "West", this );

    if( m_hide )
    {
	m_up->hide();
	m_north->hide();
	m_east->hide();
	m_south->hide();
	m_west->hide();
    }

    // set up the projections
    Vertex badUp( m_cam.m_up );
    Vertex eye( m_cam.m_eye );
    Vertex lookat( m_cam.m_lookAt );
    Vertex forward = lookat - eye; forward.normalize();
    Vertex mForward = -forward;
    Vertex third = forward ^ badUp; third.normalize();
    Vertex mThird = -third;
    Vertex up =  third ^ forward;
    Vertex mUp = -up;

    Vertex center;
	
    m_up->setPerspective( 90, m_near, m_far );
    m_up->setLookAt( eye.v, lookat.v, up.v );

#ifndef USE_SINGLE_WINDOW
    center = eye + up;
    m_north->setFrustum( -HCSIZE/2, HCSIZE/2, -HCSIZE/2, 0, m_near, m_far );
    m_north->setLookAt( eye.v, center.v, mForward.v );

    center = eye + third;
    m_east->setFrustum( -HCSIZE/2, 0, -HCSIZE/2, HCSIZE/2, m_near, m_far );
    m_east->setLookAt( eye.v, center.v, up.v );

    center = eye + mUp;
    m_south->setFrustum( -HCSIZE/2, HCSIZE/2, 0, HCSIZE/2, m_near, m_far );
    m_south->setLookAt( eye.v, center.v, forward.v );

    center = eye + mThird;
    m_west->setFrustum( 0, HCSIZE/2, -HCSIZE/2, HCSIZE/2, m_near, m_far );
    m_west->setLookAt( eye.v, center.v, up.v );
#endif    

    if( m_hide )
    {
	m_up->setCurrent();
	draw();
#ifndef USE_SINGLE_WINDOW	
	m_north->setCurrent();
	draw();
	m_east->setCurrent();
	draw();
	m_south->setCurrent();
	draw();
	m_west->setCurrent();
	draw();
#endif	
	compute();
    }
    else
	glutMainLoop();
}


void HemiCube::draw()
{
    glClearColor( 0, 0, 0, 1.0 );
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glColor3f( 0, 1, 0 );
    foreach( i, m_sources )
    {
	Polygon &p = *m_sources[i];	
#ifdef USE_ITEM_BUFFER
	GLubyte c[3];
	int label = p.label();
	convert( c, label );
	glColor3ubv( c );
#endif	
	DArray<Vertex> &verts = p.getVerts();
	glBegin( GL_POLYGON );
	foreach( j, verts )
	    glVertex3dv( verts[j]->v );
	glEnd();
    }

#if 0  // this was from before I blurred the difference between src and occ    
    glColor3f( 1, 0, 0 );
    foreach( i, m_occluders )
    {
	Polygon &p = m_occluders[i];
	DArray<Vertex> &verts = p.getVerts();
	glBegin( GL_POLYGON );
	foreach( j, verts )
	    glVertex3dv( verts[j]->v );
	glEnd();
    }
#endif    
}


void HemiCube::compute()
{
    const double d = 2./m_size;
    const double deltaA = d*d;
    double FF = 0;

#ifdef USE_ITEM_BUFFER
    // set up array to hold form factors
    double *ffBuffer = new double[m_sources.length()];
    memset( ffBuffer, 0, m_sources.length()*sizeof(double) );
#endif // USE_ITEM_BUFFER
    
    unsigned char *buffer = new unsigned char[3*m_size*m_size];
    m_up->setCurrent();
    glReadBuffer( GL_FRONT );
#ifdef USE_ITEM_BUFFER    
    glReadPixels( 0, 0, m_size, m_size, GL_RGB, GL_UNSIGNED_BYTE, buffer );
#else
    glReadPixels( 0, 0, m_size, m_size, GL_GREEN, GL_UNSIGNED_BYTE, buffer );
#endif // USE_ITEM_BUFFER

    for( int j=0; j<m_size; j++ )
    {
	int a = j*m_size;
	double y = d*(m_size/2 - j);
	for( int i=0; i<m_size; i++ )
	{
#ifdef USE_ITEM_BUFFER
	    unsigned char *c = buffer + 3*(i+a);
	    int index;
	    convert( index, c );
	    if( index == -1 )
		continue;
	    if( index < 0 || index >= m_sources.length() )
	    {
		cerr << "Bad label read from buffer:  " << index << endl;
		exit( 1 );
	    }
	    double &ff = ffBuffer[index];
#else	    
	    unsigned char c = buffer[i+a];
	    if( c <= 128 )
		continue;
	    double &ff = FF;
#endif // USE_ITEM_BUFFER
	    double x = d*(m_size/2 - i);
	    double m = x*x + y*y + 1;
	    ff += 1./(m*m);
	}
    }

#ifndef USE_SINGLE_WINDOW    
    BWWindow *windows[4];
    windows[0] = m_north;
    windows[1] = m_east;
    windows[2] = m_south;
    windows[3] = m_west;
    for( int k=0; k<4; k++ )
    {
	windows[k]->setCurrent();
	glReadBuffer( GL_FRONT );
#ifdef USE_ITEM_BUFFER	
	glReadPixels( 0, 0, m_size, m_size/2, GL_RGB, GL_UNSIGNED_BYTE, 
	    buffer );
#else	
	glReadPixels( 0, 0, m_size, m_size/2, GL_GREEN, GL_UNSIGNED_BYTE, 
	    buffer );
#endif	

	for( int j=0; j<m_size/2; j++ )
	{
	    int a = j*m_size;
	    double z = d*j;
	    for( int i=0; i<m_size; i++ )
	    {
#ifdef USE_ITEM_BUFFER		
		unsigned char *c = buffer + 3*(i+a);
		int index;
		convert( index, c );
		if( index < 0 || index >= m_sources.length() )
		{
		    cerr << "Bad label read from buffer!\n";
		    exit( 1 );
		}
		double &ff = ffBuffer[index];
#else		
	    unsigned char c = buffer[i+a];
		if( c <= 128 )
		    continue;
		double &ff = FF;
#endif		
		double y = d*(m_size/2 - i);
		double m = y*y + z*z + 1;
		ff += z/(m*m);
	    }
	}
    }
#endif // USE_SINGLE_WINDOW    

#ifdef USE_ITEM_BUFFER
    cout << " Hemi-Cube form factors:\n";
    for( int i=0; i<m_sources.length(); i++ )
    {
	int label = m_sources[i]->label();
	cout <<  "  Source " << label << ":  "
	     <<	ffBuffer[label]*deltaA/M_PI << endl;
	FF += ffBuffer[label];
    }
    delete ffBuffer;
#endif    
    
    FF *= deltaA / M_PI;
    cout << " Hemi-Cube total form factor: " << FF << '\n';
    delete buffer;

    if( !m_hide )
	exit( 0 );
}


// convert the label to colour
void convert( GLubyte *c, int label )
{
    label++; // because we clear to black, so must avoid this colour
    c[0] = (GLubyte) (label & 0xFF);
    c[1] = (GLubyte) ((label & 0xFF00) >> 8);
    c[2] = (GLubyte) ((label & 0xFF0000) >> 16);

}

void convert( int &label, GLubyte *c )
{
    label = ((int) c[0]) | (((int) c[1]) << 8) | (((int)c[2]) << 16);
    label--;
}
    
