#include <string.h>
#include <math.h>
#include <fstream.h>
#include <stdlib.h>
#include <libtk/dasort.h>
#include "jfobj.h"
#include "base.h"

class Edge
{
public:
    int v1;
    int v2;
    Edge( int x, int y ) { v1=x; v2=y; }
};



DArray<Polygon> 
JFFile::read( const char *f )
{
    ifstream in( f );

    if( in.bad() )
    {
	cerr << "Bad file: " << f << '\n';
	exit( 1 );
    }
    return read( in );
}


DArray<Polygon> 
JFFile::read( istream &in )
{
    DArray<Vertex> vertices;
    DArray<Edge> edges;
    DArray<Polygon> polygons;

    int numVertices;
    in >> numVertices;
    // cout << "Reading " << numVertices << " vertices.\n";
    for( int i=0; i<numVertices; i++ )
    {
	double v[3];

	in >> v[0] >> v[1] >> v[2];
	vertices.add( new Vertex(v) );
    }

    int numEdges;
    in >> numEdges;
    // cout << "Reading " << numEdges << " edges.\n";
    for( int i=0; i<numEdges; i++ )
    {
	int v1, v2;

	in >> v1 >> v2;
	if( v1 < 0 || v1 >= numVertices ||
	    v2 < 0 || v2 >= numVertices )
	{
	    cerr << "Edge index exceeds number of vertices!\n";
	    exit( 1 );
	}
	edges.add( new Edge(v1,v2) );
    }

    int numPolygons;
    in >> numPolygons;
    // cout << "Reading " << numPolygons << " polygons\n";
    for( int i=0; i<numPolygons; i++ )
    {
	int e;
	in >> e;
	DArray<Vertex> polyVerts( e );
	double garbage;
	for( int j=0; j< e; j++ )
	{
	    int index;
	    in >> index;
	    if( index < 0 || index >= numEdges )
	    {
		cerr << "Polygon index exceeds number of edges: "
		     << index << endl;
		exit( 1 );
	    }
	    polyVerts.add( vertices[edges[index]->v1] ); 
	}
	in >> garbage >> garbage >> garbage;
	polygons.add( new Polygon(polyVerts,i) );
	polyVerts.clear(); // don't unallocate contents
    }
    edges.clear( 1 ); // unallocate
    vertices.clear( 1 ); //unallocate
    return polygons;
}
