#include <string.h>
#include <math.h>
#include <fstream.h>
#include <stdlib.h>
#include <libtk/dasort.h>
#include "nfobj.h"
#include "base.h"


DArray<Polygon> 
NFFile::read( const char *f )
{
    ifstream in( f );

    if( in.bad() )
    {
	cerr << "Bad file: " << f << '\n';
	exit( 1 );
    }
    return read( in );
}


DArray<Polygon> 
NFFile::read( istream &in )
{
    DArray<Vertex> vertices;
    DArray<Polygon> polygons;

    int numVertices;
    in >> numVertices;
    // cout << "Reading " << numVertices << " vertices.\n";
    for( int i=0; i<numVertices; i++ )
    {
	double v[3];

	in >> v[0] >> v[1] >> v[2];
	vertices.add( new Vertex(v) );
    }

    int numPolygons;
    in >> numPolygons;
    // cout << "Reading " << numPolygons << " polygons\n";
    for( int i=0; i<numPolygons; i++ )
    {
	int numV;
	in >> numV;
	DArray<Vertex> polyVerts( numV );
	for( int j=0; j<numV; j++ )
	{
	    int index;
	    in >> index;
	    if( index < 0 || index >= numVertices )
	    {
		cerr << "Polygon vertex-index exceeds number of vertices: "
		     << index << endl;
		exit( 1 );
	    }
	    polyVerts.add( vertices[index] ); 
	}
	polygons.add( new Polygon(polyVerts,i) );
	polyVerts.clear(); // don't unallocate contents
    }
    vertices.clear( 1 ); //unallocate
    return polygons;
}
